with Pilas, Colas;

package body Contenedores.B is
	package Cola_Nodos is new Colas(Info_Camino);
	use Cola_Nodos;

	package Camino_BSqueda is new Pilas(Info_Camino);
	use Camino_BSqueda;

	--------------------------------------
	-- Operaciones centradas en el nodo --
	--------------------------------------
	function Buscar(Nodo: in Nodo_B; E: TElemento) return Natural is
		-- Busca E en un nodo del rbol B, devolviendo su posicin
		-- o el Hijos por el quedebera continuarse la bsqueda
		Inf: Natural := 1;
		Sup: Natural := Nodo.Num_Elementos;
		Med: Natural;
	begin
		-- Se realiza una bsqueda dicotmica
		while Inf <= Sup loop
			Med := (Inf + Sup) / 2;
			if Nodo.Elementos(Med) = E then
				return Med; -- Posicin de E
			else
				if E < Nodo.Elementos(Med) then
					Sup := Med - 1;
				else
					Inf := Med + 1;
				end if;
			end if;
		end loop;
		return Inf - 1; -- Hijos en el que podra estar E
    end Buscar;

	procedure Insertar(
			Nodo: in out Nodo_B;
			E   : in TElemento;
			Pos : in Natural;
			Hijo: in  Natural) is
		-- Inserta la pareja (Hijo, E) en la posicin de Nodo indicada
		-- por Pos
	begin
		Nodo.Hijos(Pos + 1 .. Nodo.Num_Elementos + 1) :=
			Nodo.Hijos(Pos .. Nodo.Num_Elementos);
		Nodo.Elementos(Pos + 1 .. Nodo.Num_Elementos + 1) :=
			Nodo.Elementos(Pos .. Nodo.Num_Elementos);

		Nodo.Hijos(Pos)    := Hijo;
		Nodo.Elementos(Pos):= E;
		Nodo.Num_Elementos := Nodo.Num_Elementos + 1;
	end Insertar;

	procedure Extraer(
			Nodo: in out Nodo_B;
			Pos : in Natural) is
		-- Extrae un Hijos y un elemento en la posicin de Nodo indicada
		-- por Pos
	begin
		Nodo.Num_Elementos := Nodo.Num_Elementos - 1;
		Nodo.Hijos(Pos .. Nodo.Num_Elementos) :=
			Nodo.Hijos(Pos + 1 .. Nodo.Num_Elementos + 1);
		Nodo.Elementos(Pos .. Nodo.Num_Elementos) :=
			Nodo.Elementos(Pos + 1 .. Nodo.Num_Elementos + 1);
	end Extraer;
	----------------------------------------
	-- Operaciones de gestin de ficheros --
	----------------------------------------
	procedure Tomar_PGina(C: in out Rbol_B; Dir: out Natural) is
		Aux: Nodo_B;
	begin
		if C.Libre = 0 then
			Dir := Natural(Size(C.PGinas)) + 1;
		else
			Dir := C.Libre;
			Read(C.PGinas, Aux, Fichero_PGinas.Count(Dir));
			C.Libre := Aux.Hijos(0);
			Write(C.Info, C.Libre, 2);
		end if;
	end Tomar_PGina;

	procedure Liberar_PGina(C: in out Rbol_B; PGina: in out Nodo_B) is
	begin
		PGina.Hijos(0) := C.Libre;
		Write(C.PGinas, PGina, Fichero_PGinas.Count(PGina.PosiciN));
		C.Libre := PGina.PosiciN;
		Write(C.Info, C.Libre, 2);
	end Liberar_PGina;

	----------------------------------
	-- Operaciones rbol multirrama --
	----------------------------------
	procedure Buscar(
			Rbol: in Rbol_B;
			Dato: in TElemento;
			Encontrado: out Boolean;
			Camino    : out Pila
		    ) is
		-- Localiza un elemento en el rbol B, y guarda en una pila
		-- el camino hasta el nodo en el que se encuentra (o en el
		-- que debera encontrarse, si no est)
		Dir, Pos   : Natural;
		Nodo_Actual: Nodo_B;
		Nodo_Camino: Info_Camino;
	begin
		-- Se empieza a buscar en la raz
		Dir := Rbol.RaZ;
		while Dir /= 0 loop
			-- Acceder al nodo actual y buscar el valor en l
			Read(Rbol.PGinas, Nodo_Actual, Fichero_PGinas.Count(Dir));
			Pos := Buscar(Nodo_Actual, Dato);
			Nodo_Camino.Nodo := Nodo_Actual;
			Nodo_Camino.Pos  := Pos;
			Apilar(Camino, Nodo_Camino);
			if Pos > 0 and then Nodo_Actual.Elementos(Pos) = Dato then
				Encontrado := True;
				return;
			end if;
			-- La bsqueda en el rbol continuar en el nodo indicado por la 
			-- bsqueda en el nodo actual
			Dir := Nodo_Actual.Hijos(Pos);
		end loop;
		Encontrado := False;
	end Buscar;

	-----------------------------
	-- Operaciones del rbol B --
	-----------------------------
	procedure ParticiN_1_2(
			C:in out Rbol_B;
			Nodo: in out Nodo_B;
			Elementos_Extra: out TElemento;
			Dir_Nuevo: out Natural) is
		Nuevo: Nodo_B;
		Sup, Inf  : Natural;
	begin
		Tomar_PGina(C, Dir_Nuevo);
		Nuevo.PosiciN := Dir_Nuevo;
		Sup := (Orden + 1) / 2; -- Redondeo superior de Orden
		Nuevo.Hijos(0 .. Nodo.Num_Elementos - Sup) := Nodo.Hijos(Sup .. Nodo.Num_Elementos);
		Nuevo.Elementos(1 .. Nodo.Num_Elementos - Sup) := Nodo.Elementos(Sup + 1 .. Nodo.Num_Elementos); 
		Nuevo.Num_Elementos := Nodo.Num_Elementos - Sup;
		Nodo.Num_Elementos  := Sup - 1;
		Elementos_Extra := Nodo.Elementos(Sup);
		Write(C.PGinas, Nodo, Fichero_PGinas.Count(Nodo.PosiciN));
		Write(C.PGinas, Nuevo, Fichero_PGinas.Count(Nuevo.PosiciN));
	end ParticiN_1_2;

	procedure ParticiN_2_3(
			C: in out Rbol_B;
			Padre: in out Nodo_B;
			Pos_Izq: in Natural;
			Izq: in out Nodo_B;
			Der: in out Nodo_B
			) is
		Repartir: Natural := Izq.Num_Elementos + Der.Num_Elementos - 1;
		Reg     : Nodo_B;
		Ncizq   : Natural := (Repartir) / 3;
		Ncreg   : Natural := (Repartir + 1) / 3;
		Ncder   : Natural := (Repartir + 2) / 3;
		Ant_Ncder: Natural := Der.Num_Elementos;
		Ant_Ncizq: Natural := Izq.Num_Elementos;
		Posl     : Natural := Ant_Ncizq - Ncizq;
		Ncpas    : Natural;
	begin
		-- Se inserta en el padre una nueva clave y la nueva direccion
		Tomar_PGina(C, Reg.PosiciN);
		Insertar(Padre, Izq.Elementos(Ncizq + 1), Pos_Izq + 1, Reg.PosiciN);

		-- Pasamos datos de izq a reg
		Reg.Num_Elementos := Ncreg;
		Reg.Elementos(1 .. Ant_Ncizq - Ncizq - 1) :=
			Izq.Elementos(Ncizq + 2 .. Ant_Ncizq);
		Reg.Hijos(0 .. Ant_Ncizq - Ncizq - 1) :=
			Izq.Hijos(Ncizq + 1 .. Ant_Ncizq);
		Izq.Num_Elementos := Ncizq;
		Write(C.PGinas, Izq, Fichero_PGinas.Count(Izq.PosiciN));

		-- Pasamos el dato del padre a la posicin correspondiente de Reg
		Reg.Elementos(Ant_Ncizq - Ncizq) := Padre.Elementos(Pos_Izq + 2);

		-- Pasamos elementos de Der a Reg
		Reg.Elementos(Posl + 1 .. Ncreg) :=
			Der.Elementos(1 .. Ncreg - Posl);
		Reg.Hijos(Posl .. Ncreg) :=
			Der.Hijos(0 .. Ncreg - Posl);
		Write(C.PGinas, Reg, Fichero_PGinas.Count(Reg.PosiciN));

		-- Pasamos al padre el valor correspondiente y compactamos Der
		Ncpas := Ant_Ncder - Ncder;
		Padre.Elementos(Pos_Izq + 2) := Der.Elementos(Ncpas);
		Der.Elementos(1 .. Ant_Ncder - Ncpas) :=
			Der.Elementos(Ncpas + 1 .. Ant_Ncder);
		Der.Hijos(0 .. Ant_Ncder - Ncpas) :=
			Der.Hijos(Ncpas .. Ant_Ncder);
		Der.Num_Elementos := Ncder;
		Write(C.PGinas, Der, Fichero_PGinas.Count(Der.PosiciN));
	end ParticiN_2_3;

	procedure Rotar_Izq_Der
			(C: in out Rbol_B;
			Padre: in out Nodo_B; -- Nodo padre de Izq
			Pos_Izq: in Natural;  -- Posicin dentro del nodo padre del Hijos a Izq
			Izq: in out Nodo_B;   -- Nodo sobrecargado
			Der: in out Nodo_B    -- Nodo hermano inmediato derecho de Izq
			) is
		N_Der, N_Izq, N_Pas: Natural;
	begin
		-- Desplazamiento de Der para hacer hueco
		N_Izq := (Izq.Num_Elementos + Der.Num_Elementos) / 2;
		N_Der := (Izq.Num_Elementos + Der.Num_Elementos) - N_Izq;
		N_Pas :=  N_Der - Der.Num_Elementos;

		Der.Elementos(N_Pas + 1 .. Der.Num_Elementos + N_Pas) :=
			Der.Elementos(1 .. Der.Num_Elementos);

		Der.Hijos(N_Pas .. Der.Num_Elementos + N_Pas) :=
			Der.Hijos(0 .. Der.Num_Elementos);

		-- Rellenar el nodo Der
		Der.Elementos(N_Pas) := Padre.Elementos(Pos_Izq + 1);
		Der.Elementos(1 .. N_Pas - 1) :=
			Izq.Elementos(N_Izq + 2 .. Izq.Num_Elementos);
		Der.Hijos(1 .. N_Pas - 1) :=
			Izq.Hijos(N_Izq + 2 .. Izq.Num_Elementos);
		Der.Hijos(0) := Izq.Hijos(N_Izq + 1);
		Der.Num_Elementos := N_Der;
		Write(C.PGinas, Der, Fichero_PGinas.Count(Der.PosiciN));

		-- Modificar el nodo Padre
		Padre.Elementos(Pos_Izq + 1) := Izq.Elementos(N_Izq + 1);
		Write(C.PGinas, Padre, Fichero_PGinas.Count(Padre.PosiciN));
		
		-- Modificar el nodo Izq
		Izq.Num_Elementos := N_Izq;
		Write(C.PGinas, Izq, Fichero_PGinas.Count(Izq.PosiciN));
	end Rotar_Izq_Der;

	procedure Rotar_Der_Izq
			(C: in out Rbol_B;
			Padre: in out Nodo_B; -- Nodo padre de Der
			Pos_Der: in Natural;  -- Posicin dentro del nodo padre del enlace a Der
			Izq: in out Nodo_B;   -- Nodo hermano inmediato izquierdo de Der
			Der: in out Nodo_B    -- Nodo sobrecargado
			) is

		N_Der, N_Izq, N_Pas: Natural;
	begin
		N_Der := (Izq.Num_Elementos + Der.Num_Elementos) / 2;
		N_Izq := (Izq.Num_Elementos + Der.Num_Elementos) - N_Der;
		N_Pas :=  N_Izq - Izq.Num_Elementos;

		-- Rellenar el nodo Izq
		Izq.Hijos(Izq.Num_Elementos + 1) := Der.Hijos(0);
		Izq.Elementos(Izq.Num_Elementos + 1) := Padre.Elementos(Pos_Der);
		Izq.Hijos(Izq.Num_Elementos + 2 .. N_Izq) :=
			Der.Hijos(1 .. N_Izq - (Izq.Num_Elementos + 1));
		Izq.Elementos(Izq.Num_Elementos + 2 .. N_Izq) :=
			Der.Elementos(1 .. N_Izq - (Izq.Num_Elementos + 1));
		Izq.Num_Elementos := N_Izq;
		Write(C.PGinas, Izq, Fichero_PGinas.Count(Izq.PosiciN));

		-- Modificar el nodo Padre
		Padre.Elementos(Pos_Der) := Der.Elementos(Der.Num_Elementos - N_Der);
		Write(C.PGinas, Padre, Fichero_PGinas.Count(Padre.PosiciN));

		-- Desplazamiento de Der para rellenar el hueco dejado
		Der.Elementos(1 .. Der.Num_Elementos - N_Pas) :=
			Der.Elementos(N_Pas + 1 .. Der.Num_Elementos);

		Der.Hijos(0 .. Der.Num_Elementos - N_Pas) :=
			Der.Hijos(N_Pas .. Der.Num_Elementos);
		Der.Num_Elementos := N_Der;
		Write(C.PGinas, Der, Fichero_PGinas.Count(Der.PosiciN));
	end Rotar_Der_Izq;

	procedure Insertar(C: in out Rbol_B; E: in TElemento) is
		Encontrado: Boolean;
		Camino    : Pila;
		Elemento_Insertar : TElemento := E;
		Hijo_Insertar: Natural := 0;
		Nodo_Camino: Info_Camino;
		Nueva_RaZ : Nodo_B;
		Izq, Der, Padre, Actual : Nodo_B;
		Pos: Natural;
	begin
		-- El primer paso de la insercin es intentar encontrar el elemento
		Buscar(C, E, Encontrado, Camino);
		if not Encontrado then
			C.Num_Elementos := C.Num_Elementos + 1;
			Write(C.Info, C.Num_Elementos, 3);
			if not Es_VacA(Camino) then
				-- Se inserta en el nodo que est al final del camino
				Nodo_Camino := Examinar(Camino);
				Actual := Nodo_Camino.Nodo;
				Pos    := Nodo_Camino.Pos;
				Desapilar(Camino);
				Insertar(Actual, Elemento_Insertar, Pos + 1, Hijo_Insertar);
				if Actual.Num_Elementos < Orden then
					Write(C.PGinas, Actual, Fichero_PGinas.Count(Actual.PosiciN));
					return;
				end if;
				
				-- Si el nodo se sobrecarga, se intenta partir o rotar, si se
				-- parte, ola sobrecarga puede propagarse hasta la raz
				while not Es_VacA(Camino) loop
					Nodo_Camino := Examinar(Camino);
					Padre := Nodo_Camino.Nodo;
					Pos   := Nodo_Camino.Pos;
					Desapilar(Camino);

					-- Primero se intenta rotar
					if Pos > 0 then
						Read(C.PGinas, Izq, Fichero_PGinas.Count(Padre.Hijos(Pos - 1)));
						if Izq.Num_Elementos < Orden - 1 then
							Rotar_Der_Izq(C, Padre, Pos, Izq, Actual);
							return;
						end if;
					end if;

					if Pos < Padre.Num_Elementos then
						Read(C.PGinas, Der, Fichero_PGinas.Count(Padre.Hijos(Pos + 1)));
						if Der.Num_Elementos < Orden - 1 then
							Rotar_Izq_Der(C, Padre, Pos, Actual, Der);
							return;
						end if;
					end if;

					-- Lo segundo que se intenta es una particin-2/3
					if Pos = 0 then
						ParticiN_2_3(C, Padre, Pos, Actual, Der);
					else
						ParticiN_2_3(C, Padre, Pos - 1, Izq, Actual);
					end if;

					if Padre.Num_Elementos < Orden then
						Write(C.PGinas, Padre, Fichero_PGinas.Count(Padre.PosiciN));
						return;
					end if;
	
					Actual := Padre;
				end loop;

				-- La particin-1/2 se aplica a la raz
				ParticiN_1_2(C, Actual, Elemento_Insertar, Hijo_Insertar);
			end if;

			-- Si se llega aqu, hay que crear una nueva raz
			Tomar_PGina(C, Nueva_RaZ.PosiciN);
			Nueva_RaZ.Hijos(0) := C.RaZ;
			Nueva_RaZ.Elementos(1) := Elemento_Insertar;
			Nueva_RaZ.Hijos(1) := Hijo_Insertar;
			Nueva_RaZ.Num_Elementos := 1;
			Write(C.PGinas, Nueva_RaZ, Fichero_PGinas.Count(Nueva_RaZ.PosiciN));
			C.RaZ := Nueva_RaZ.PosiciN;
			Write(C.Info, C.RaZ, 1);
		end if;
	end Insertar;

	procedure RecombinaciN_2_1(
			C:in out Rbol_B;
			Nodo: in out Nodo_B; -- Nodo izquierdo en la recombinacin
			Pos_Nodo: in Natural;-- Posicin  de Nodo en Padre	
			Padre: in out Nodo_B;-- Padre de los nodos a recombinar
			Der: in out Nodo_B   -- Nodo derecho en la recombinacin
			) is
	begin
		-- Se pasa a Nodo la Elementos de Padre que lo separa de Der
		Nodo.Elementos(Nodo.Num_Elementos + 1) := Padre.Elementos(Pos_Nodo + 1);
		Nodo.Num_Elementos := Nodo.Num_Elementos + 1;
		Extraer(Padre, Pos_Nodo + 1);

		-- Se pasa a Nodo el contenido de Der, que est bajo mnimo
		Nodo.Elementos
			(Nodo.Num_Elementos + 1 .. Nodo.Num_Elementos + Der.Num_Elementos) :=
			Der.Elementos(1 .. Der.Num_Elementos);
		Nodo.Hijos
			(Nodo.Num_Elementos .. Nodo.Num_Elementos + Der.Num_Elementos) :=
			Der.Hijos(0 .. Der.Num_Elementos);
		Nodo.Num_Elementos := Nodo.Num_Elementos + Der.Num_Elementos;
		Write(C.PGinas, Nodo, Fichero_PGinas.Count(Nodo.PosiciN));

		-- Se libera Der
		Liberar_PGina(C, Der);
	end RecombinaciN_2_1;

	procedure RecombinaciN_3_2(
			C: in out Rbol_B;
			Padre: in out Nodo_B; -- Padre de los nodos a recompbinar
			Pos_Reg: in Natural;  -- Posicin de Reg en Padre
			Izq: in out Nodo_B;   -- Nodo izquierdo a recombinar
			Reg: in out Nodo_B;   -- Nodo central a recombinar
			Der: in out Nodo_B    -- Nodo derecho a recombinar
			) is
		Ncder, Ncizq, Ncpas: Natural;
	begin
		Ncder := (Reg.Num_Elementos + Der.Num_Elementos + Izq.Num_Elementos + 1) / 2; 
		Ncizq := (Reg.Num_Elementos + Der.Num_Elementos + Izq.Num_Elementos + 1) - Ncder; 

		-- Se pasan elementos y enlaces a Izq
		Izq.Elementos(Izq.Num_Elementos + 1):= Padre.Elementos(Pos_Reg);
		Izq.Elementos(Izq.Num_Elementos + 2 .. Ncizq) :=
			Reg.Elementos(1 .. Ncizq - Izq.Num_Elementos - 1);
		Izq.Hijos(Izq.Num_Elementos + 1 .. Ncizq) :=
			Reg.Hijos(0 .. Ncizq - Izq.Num_Elementos - 1);
		Izq.Num_Elementos := Ncizq;
		Write(C.PGinas, Izq, Fichero_PGinas.Count(Izq.PosiciN));

		-- Desplazamiento de Der para hacer hueco
		Ncpas := Ncder - Der.Num_Elementos;
		Der.Elementos(Ncpas + 1.. Der.Num_Elementos + Ncpas) :=
			Der.Elementos(1 .. Der.Num_Elementos);
		Der.Hijos(Ncpas .. Der.Num_Elementos + Ncpas) :=
			Der.Hijos(0 .. Der.Num_Elementos);

		-- Se pasan elementos y enlaces a Der
		Der.Elementos(Ncpas) := Padre.Elementos(Pos_Reg + 1);
		Der.Elementos(1 .. Ncpas - 1) :=
			Reg.Elementos(Reg.Num_Elementos - Ncpas + 2 .. Reg.Num_Elementos);
		Der.Hijos(0 .. Ncpas - 1) :=
			Reg.Hijos(Reg.Num_Elementos - Ncpas + 1 .. Reg.Num_Elementos);
		Der.Num_Elementos := Ncder;
		Write(C.PGinas, Der, Fichero_PGinas.Count(Der.PosiciN));

		-- Modificar El Nodo Padre
		Extraer(Padre, Pos_Reg);
		Padre.Elementos(Pos_Reg) := Reg.Elementos(Reg.Num_Elementos - Ncpas + 1);
		Liberar_PGina(C, Reg);
	end RecombinaciN_3_2;

	procedure Extraer(C: in out Rbol_B; E: in TElemento) is
		Encontrado : Boolean := False;
		Camino     : Pila;
		Nodo_Camino: Info_Camino;
		Camino_Suc: Cola;
		Actual: Nodo_B;
		Pos   : Natural;
		Dir   : Natural;
		Padre, Der, Izq: Nodo_B;
	begin
		Buscar(C, E, Encontrado, Camino);
		if Encontrado then
			C.Num_Elementos := C.Num_Elementos - 1;
			Write(C.Info, C.Num_Elementos, 3);

			Nodo_Camino := Examinar(Camino);
			Actual := Nodo_Camino.Nodo;
			Pos    := Nodo_Camino.Pos;
			if Actual.Hijos(0) /= 0 then
				-- Extraccin desde un nodo no hoja
				-- Hay que buscar el sucesor y cambiarlo
				Dir := Actual.Hijos(Pos);
				loop
					-- Descendemos por la rama izquierda del subrbol del sucesor
					Read(C.PGinas, Actual, Fichero_PGinas.Count(Dir));
					Dir := Actual.Hijos(0);
					if Dir = 0 then
						Pos := 1;
					else
						Pos := 0;
					end if;

					-- Guardamos el camino al sucesor
					Insertar(Camino_Suc,(Actual, Pos));

					exit when Dir = 0;
				end loop;

				Nodo_Camino := Examinar(Camino);
				Desapilar(Camino);
				-- Se sustituye por el sucesor
				Nodo_Camino.Nodo.Elementos(Nodo_Camino.Pos) := Actual.Elementos(1);
 				Write(C.PGinas, Nodo_Camino.Nodo, Fichero_PGinas.Count(Nodo_Camino.Nodo.PosiciN));

				Apilar(Camino, Nodo_Camino);
				while not Es_VacA(Camino_Suc) loop
					Apilar(Camino, Examinar(Camino_Suc));
					Extraer(Camino_Suc);
				end loop;
			end if;

			Extraer(Actual, Pos);
			while (Actual.Num_Elementos < MNimo) and (Actual.PosiciN /= C.RaZ) loop
				Desapilar(Camino);

				-- Sacamos el padre de la pila
				Nodo_Camino := Examinar(Camino);
				Padre := Nodo_Camino.Nodo;
				Pos   := Nodo_Camino.Pos;

				if Pos < Padre.Num_Elementos then
					-- Tiene hermano derecho
					Read(C.PGinas, Der, Fichero_PGinas.Count(Padre.Hijos(Pos + 1)));
					if Der.Num_Elementos > MNimo then
						Rotar_Der_Izq(C, Padre, Pos + 1, Actual, Der);
						return;
					end if;
				end if;

				if Pos > 0 then
					-- Tiene hermano izquierdo
					Read(C.PGinas, Izq, Fichero_PGinas.Count(Padre.Hijos(Pos - 1)));
					if Izq.Num_Elementos > MNimo then
						Rotar_Izq_Der(C, Padre, Pos - 1, Izq, Actual);
						return;
					end if;
				end if;

				if (Pos > 0) and then (Pos < Padre.Num_Elementos) then
					-- Tiene hermanos izquierdo y derecho
					RecombinaciN_3_2(C, Padre, Pos, Izq, Actual, Der);
				elsif (Pos > 0) then
					-- Tiene hermano izquierdo pero no derecho
					RecombinaciN_2_1(C, Izq, Pos - 1, Padre, Actual);
				else
					-- Tiene hermano derecho pero no izquierdo				
					RecombinaciN_2_1(C, Actual, Pos, Padre, Der);
				end if;
				Actual := Padre;
			end loop;

			if (Actual.Num_Elementos > 0) then
				-- Se escribe el nodo
				Write(C.PGinas, Actual, Fichero_PGinas.Count(Actual.PosiciN));
			else
				-- La raz se ha quedado vaca
				C.RaZ := Actual.Hijos(0);
				Liberar_PGina(C, Actual);
				Write(C.Info, C.RaZ, 1);
			end if;
		end if;
	end Extraer;

	-----------------------------------------
	-- operaciones pblicas del contenedor --
	-----------------------------------------
	procedure Vaciar(C: in out Rbol_B) is
	begin
		Crear(C, C.Ruta);
	end Vaciar;

	function Est(C: in Rbol_B; E: in TElemento) return Boolean is
		Encontrado: Boolean;
		Camino    : Pila;
	begin
		Buscar(C, E, Encontrado, Camino);
		return Encontrado;
	end Est;
		
	function TamaO(C: in Rbol_B) return Natural is
	begin
		return C.Num_Elementos;
	end TamaO;

	
	procedure Abrir(C: in out Rbol_B; Ruta: in Unbounded_String) is
	begin
		if Is_Open(C.Info) then
			Cerrar(C);
		end if;

		-- Se carga la informacin del tad
		Open(C.Info,Inout_File,To_String(Ruta & "Info.b"));
		Read(C.Info, C.RaZ);
		Read(C.Info, C.Libre);
		Read(C.Info, C.Num_Elementos);
		C.Ruta := Ruta;

		-- Se abre el fichero de datos
		Open(C.PGinas,Inout_File,To_String(Ruta & "Datos.b"));
	end Abrir;

	procedure Crear(C: in out Rbol_B; Ruta: in Unbounded_String) is
	begin
		if Is_Open(C.Info) then
			Cerrar(C);
		end if;

		-- Se inicializa el fichero de informacin auxiliar
		Create(C.Info,Name => To_String(Ruta & "Info.b"));
		Write(C.Info,0); -- Raz (nula)
		Write(C.Info,0); -- No hay nodos liberados
		Write(C.Info,0); -- N total de elementos almacenados

		-- Se inicializa la informacin interna del tad
		C.RaZ := 0;
		C.Libre:= 0;
		C.Num_Elementos := 0;
		C.Ruta := Ruta;

		-- Se crea el fichero de datos
		Create(C.PGinas,Name => To_String(Ruta & "Datos.b"));
	end Crear;

	procedure Cerrar(C: in out Rbol_B) is
	begin
		-- Se vuelca la informacin interna en el fichero de informacin
		Reset(C.Info);
		Write(C.Info, C.RaZ);
		Write(C.Info, C.Libre);
		Write(C.Info, C.Num_Elementos);

		-- Se cierran los ficheros
		Close(C.Info);
		Close(C.PGinas);
	end Cerrar;
end Contenedores.B;
