with Ada.Strings.Unbounded, Direct_Io;
use  Ada.Strings.Unbounded;

generic
	Orden: Positive; -- Orden del rbol B
	with function "<"(A, B: in TElemento) return Boolean is <>;
package Contenedores.B is
	type Rbol_B is new Contenedor with private;

	procedure Insertar(C: in out Rbol_B; E: in TElemento);
	procedure Extraer(C: in out Rbol_B; E: in TElemento);
	procedure Vaciar(C: in out Rbol_B);
	function Est(C: in Rbol_B; E: in TElemento) return Boolean;
	function TamaO(C: in Rbol_B) return Natural;

	---------------------------------------------
	-- Operaciones para gestionar los ficheros --
	-- asociados al rbol B                    --
	---------------------------------------------
	procedure Abrir(C: in out Rbol_B; Ruta: in Unbounded_String);
		-- Abre un rbol B existente en la ubicacin indicada por Ruta
	procedure Crear(C: in out Rbol_B; Ruta: in Unbounded_String);
		-- Crea un rbol B en la ubicacin indicada por Ruta
	procedure Cerrar(C: in out Rbol_B);
		-- Cierra el fichero asociado al rbol B

private
	-- Estructura nodal del rbol B
	type Almacen is array(1 .. Orden) of TElemento;
	type Lista_Hijos is array(0 .. Orden) of Natural;
		-- Hay una clave y un enlace de ms para manejar la sobrecarga
		-- antes de rotar o partir
	type Nodo_B is record
		PosiciN : Positive; -- Posicin del nodo en el fichero 
		Elementos: Almacen;
		Hijos    : Lista_Hijos := (others => 0);
		Num_Elementos: Natural;
	end record;

	-- Instanciacin para el fichero donde se ubica el rbol B
	package Fichero_PGinas is new Direct_Io(Nodo_B);
	use Fichero_PGinas;

	-- Instanciacin para el fichero auxiliar de informacin
	-- sobre el rbol B
	package Fichero_Info is new Direct_Io(Natural);
	use Fichero_Info;

	-- Estructuras para representar el camino de bsqueda
	type Info_Camino is record
		Nodo: Nodo_B;
		Pos : Natural;
	end record;

	-- Estructura interna del rbol B
	type Rbol_B is new Contenedor with record
		Ruta   : Unbounded_String;	-- Ubicacin de los ficheros
		Info   : Fichero_Info.File_Type;   	-- Fichero de informacin auxiliar
		PGinas: Fichero_PGinas.File_Type;	-- Fichero de datos
		RaZ   : Natural;  -- Raz del rbol
		Libre  : Natural;  -- Primero de la lista de nodos liberados reutilizables
		Num_Elementos: Natural;  -- Nmero de elementos almacenados
    end record;

	MNimo: constant Natural := (Orden + 1) / 2 - 1;
end Contenedores.B;
