#ifndef INC_ARBOL_B
#define INC_ARBOL_B
#include "multirrama.h"
#include <queue>
   using namespace std;
/**
 * Implementacin de un contenedor usando un rbol B
 */
template <class T, int Orden>
   class ArbolB:public Multirrama<T,Orden>{
   protected:
      typedef NodoMR<T,Orden> Nodo; //alias para los nodosMR
      typedef infoPila<T,Orden> iPila;
      void particion_1_2(Nodo &nodo, T &clavepadre, int &dirNuevo);
      void particion_2_3(Nodo &padre, int posizq, Nodo &izq, Nodo &der);
      void rotacionizqder(Nodo &padre, int posizq, Nodo &izq, Nodo &der);
      void rotacionderizq(Nodo &padre, int posizq, Nodo &izq, Nodo &der);
      void recombinacion_2_1(Nodo &padre, int, Nodo &izq, Nodo &der);
      void recombinacion_3_2(Nodo &padre, int, Nodo &izq, Nodo &cen,Nodo &der);
      int minimoclaves;
   public:
      ArbolB(){minimoclaves=(Orden+1)/2-1;}
      virtual void insertar(const T &);
      virtual void extraer(const T &);
      virtual void crear(const char *ruta);
      virtual void abrir(const char *ruta);
      virtual void cerrar();
      virtual void vaciar();
   };
template <class T, int Orden>
   void ArbolB<T,Orden>::crear(const char *ruta){
      cerrar();
      this->nombreFichero=ruta;
      this->fichero.crear(this->nombreFichero,2);
      this->raiz=dirNula;
      this->numEle=0;
      this->fichero.adjunto(0,this->raiz);
      this->fichero.adjunto(1,this->numEle);
   };
template <class T, int Orden>
   void ArbolB<T,Orden>::abrir(const char *ruta){
      this->fichero.abrir(ruta);
      this->raiz=this->fichero.adjunto(0);
      this->numEle=this->fichero.adjunto(1);
   };
template <class T, int Orden>
   void ArbolB<T,Orden>::cerrar(){
      this->fichero.cerrar();
   };
template <class T, int Orden>
   void ArbolB<T,Orden>::vaciar(){
      this->fichero.cerrar();
      crear(this->nombreFichero.c_str());
   };
template <class T, int Orden>
   void ArbolB<T,Orden>::particion_1_2(Nodo &nodo, T &clavepadre, int &dirNuevo){
      Nodo nuevoNodo;
      int ncnuevo=Orden/2;
      int ncnodo=Orden-ncnuevo-1;
      dirNuevo=this->fichero.tomarPagina();
      nuevoNodo.direccion(dirNuevo);
      nuevoNodo.cardinal(ncnuevo);
      nuevoNodo.enlace(0)=nodo.enlace(ncnodo+1);
      //[xxxx] [   ] => [xx  ] [xx  ]
      for(int i=1; i<=nuevoNodo.cardinal(); i++){
         nuevoNodo.clave(i)=nodo.clave(ncnodo+1+i);
         nuevoNodo.enlace(i)=nodo.enlace(ncnodo+1+i);
      }
      clavepadre=nodo.clave(ncnodo+1);
      dirNuevo=nuevoNodo.direccion();
      nodo.cardinal(ncnodo);
      escribir(nodo);
      escribir(nuevoNodo);
   }
template <class T, int Orden>
   void ArbolB<T,Orden>::particion_2_3(Nodo &padre, int posizq, Nodo &izq, Nodo &der){
      int clavesRepartir=izq.cardinal()+der.cardinal()-1;
      Nodo reg;
      int ncizq =(clavesRepartir)/3;
      int ncreg =(clavesRepartir+1)/3;
      int ncder =(clavesRepartir+2)/3;
      int antncder = der.cardinal();
      int antncizq = izq.cardinal();
      //Se inserta en el padre una nueva clave y la nueva direccion
      reg.direccion(this->fichero.tomarPagina());
      padre.insertar(izq.clave(ncizq+1),reg.direccion(),posizq+1);
      //Pasamos datos de izq a reg: [xxx] [   ] => [xx ] [x  ]
      reg.cardinal(ncreg);
      reg.enlace(0)=izq.enlace(ncizq+1);
      for(int i= ncizq+2; i<=antncizq; i++) {
         reg.clave(i-ncizq-1) = izq.clave(i);
         reg.enlace(i-ncizq-1) = izq.enlace(i);
      }
      izq.cardinal(ncizq);
      //Pasamos el dato del padre a la posicin correspondiente de reg
      reg.clave(antncizq-ncizq)=padre.clave(posizq+2);
      int posl=antncizq-ncizq;
      reg.enlace(posl)=der.enlace(0); // [x  ] [yyy] => [xy ] [ yy]
      for(int i= posl+1; i<=ncreg; i++) {
         reg.clave(i) = der.clave(i-posl);
         reg.enlace(i) = der.enlace(i-posl);
      }
      int ncpas=antncder-ncder;
      //Pasamos al padre el valor correspondiente y compactamos der
      padre.clave(posizq+2)=der.clave(ncpas);
      der.enlace(0) = der.enlace(ncpas); // [ yy] => [yy ]
      for(int i= ncpas+1; i<=antncder; i++) {
         der.clave(i-ncpas) = der.clave(i);
         der.enlace(i-ncpas) = der.enlace(i);
      }
      der.cardinal(ncder);
      escribir(izq); // [xx ] [xy ] [yy ]
      escribir(reg);
      escribir(der);
   }
template <class T, int Orden>
   void ArbolB<T,Orden>::rotacionizqder(Nodo &padre, int posizq, Nodo &izq,Nodo &der){
      int clavesRepartir=izq.cardinal()+der.cardinal();
      int ncizq =(clavesRepartir)/2;
      int ncder = clavesRepartir-ncizq;
      int ncpas = ncder-der.cardinal();
      int antncder = der.cardinal();
   //Hacemos hueco en nodo der: [yy  ] => [ yy ]
      der.cardinal(ncder);
      for(int i= antncder; i>=1; i--) {
         der.clave(i+ncpas) = der.clave(i);
         der.enlace(i+ncpas) = der.enlace(i);
      }
      der.enlace(ncpas) = der.enlace(0);
   //Rellenar el nodo der: [xxxx] [ yy ] => [xxx ] [xyy ]
      der.clave(ncpas) = padre.clave(posizq+1);
      for(int i = ncizq+2; i<= izq.cardinal(); i++){
         der.clave(i-(ncizq+1)) = izq.clave(i);
         der.enlace(i-(ncizq+1)) = izq.enlace(i);
      }
      der.enlace(0) = izq.enlace(ncizq+1);
   // Modificar el nodo padre
      padre.clave(posizq+1) = izq.clave(ncizq+1);
   //Modificar el nodo izq
      izq.cardinal(ncizq);
   //Se escribe en el fichero los tres nodos
      escribir(padre);
      escribir(izq);
      escribir(der);
   }

template <class T, int Orden>
   void ArbolB<T,Orden>::rotacionderizq(Nodo &padre, int posizq, Nodo &izq,Nodo &der){
      int clavesRepartir=izq.cardinal()+der.cardinal();
      int ncder =(clavesRepartir)/2;
      int ncizq = clavesRepartir-ncder;
      int ncpas = der.cardinal()-ncder;
      int antncizq = izq.cardinal();
   //Pasamos la clave del padre y datos de der a izq
      izq.cardinal(ncizq);
      izq.clave(antncizq+1)= padre.clave(posizq+1);
      izq.enlace(antncizq+1)=der.enlace(0);
      for(int i= 1; i<ncpas; i++){ // [xx  ] [yyyy] => [xxy ] [ yyy]
         izq.clave(antncizq+1+i) = der.clave(i);
         izq.enlace(antncizq+1+i) = der.enlace(i);
      }
   //Pasamos clave al padre
      padre.clave(posizq+1)=der.clave(ncpas);
   //Quitamos hueco en der
      der.enlace(0)=der.enlace(ncpas);
      for(int i= 1; i<=ncder; i++){ // [ yyy] => [yyy ]
         der.clave(i) = der.clave(i+ncpas);
         der.enlace(i) = der.enlace(i+ncpas);
      }
      der.cardinal(ncder);
      escribir(padre);
      escribir(izq);
      escribir(der);
   }
template <class T, int Orden>
   void ArbolB<T,Orden>::recombinacion_2_1(Nodo &padre, int posizq, Nodo &izq, Nodo &der){
   	//Bajamos la clave discriminante en el padre al final del izquierdo
      int antncizq=izq.cardinal();
      izq.cardinal(izq.cardinal()+1+der.cardinal());
      izq.clave(antncizq+1)= padre.clave(posizq+1);
   	//Pasamos el enlace cero de der a izq ya que no encaja en el bucle
      izq.enlace(antncizq+1)= der.enlace(0);
   	//Pasamos el resto de claves y enlaces
      // [xx  ] [xx  ] => [xxxx] [    ]
      for(int i=1; i<=der.cardinal(); i++){
         izq.clave(antncizq+1+i)= der.clave(i);
         izq.enlace(antncizq+1+i)= der.enlace(i);
      }
   	//Quitamos del padre la clave y el enlace a der
      padre.extraer(posizq+1);
      escribir(izq);
      this->fichero.liberarPagina(der.direccion());
   }
template <class T, int Orden>
   void ArbolB<T,Orden>::recombinacion_3_2(Nodo &padre, int posReg, Nodo &izq, Nodo &reg, Nodo &der){
      int aRepartir=izq.cardinal()+reg.cardinal()+der.cardinal()+1;
      int ncder= aRepartir/2;
      int ncizq= aRepartir-ncder;
      int antncizq=izq.cardinal();
      int antncder=der.cardinal();
      //Rellenamos el hermano izquierdo
      izq.cardinal(ncizq);
      izq.clave(antncizq+1) = padre.clave(posReg);
      izq.enlace(antncizq+1) = reg.enlace(0);
      for(int i=antncizq+2; i<=ncizq ; i++){ // [xx  ] [yy  ] => [xxy ] [ y  ]
         izq.clave(i) = reg.clave(i-antncizq-1);
         izq.enlace(i) =  reg.enlace(i-antncizq-1);
      }
      //Desplazamiento del hermano derecho para hacer hueco
      der.cardinal(ncder);
      int ncpas = ncder-antncder;
      for(int i = antncder; i>=1 ; i--){ // [zz  ] => [ zz ]
         der.clave(i+ncpas) = der.clave(i);
         der.enlace(i+ncpas)= der.enlace(i);
      }
      der.enlace(ncpas)= der.enlace(0);
      der.clave(ncpas) = padre.clave(posReg+1);
      //Rellenamos el hermano derecho
      for(int i=ncpas-1; i>=1; i--){ // [ y  ] [ zz ] => [    ] [yzz ]
         der.clave(i) = reg.clave(reg.cardinal()+i-ncpas+1);
         der.enlace(i) = reg.enlace(reg.cardinal()+i-ncpas+1);
      }
      der.enlace(0) = reg.enlace(reg.cardinal()-ncpas+1);
      //modificar el nodo padre
      this->fichero.liberarPagina(reg.direccion());
      escribir(izq);
      escribir(der);
      padre.extraer(posReg);
      padre.clave(posReg) = reg.clave(reg.cardinal()-ncpas+1);
   }

template <class T, int Orden>
   void ArbolB<T,Orden>::insertar(const T &dato){
      stack<iPila> pila;
      if(buscar(dato, pila)) //No se admiten repetidos
         return;
      Nodo nodoActual;
      iPila info;
      T claveInsertar=dato;
      int enlaceInsertar= dirNula;
      this->fichero.adjunto(1,++this->numEle);
      if(!pila.empty()){ //El rbol no est vaco
         info=pila.top();
         pila.pop();
         nodoActual= info.nodo;
         int  pos= info.pos;
         nodoActual.insertar(claveInsertar,enlaceInsertar,pos+1);
         if(nodoActual.cardinal()< Orden){ //No hay problema
            escribir(nodoActual);
            return;
         }
         while(!pila.empty()){ //Arreglar sobrecarga
            Nodo der,izq;
            Nodo padre=pila.top().nodo;
            int  pos=pila.top().pos;
            pila.pop();
            if(pos > 0){ //Tiene hermano izquierdo
               izq=leer(padre.enlace(pos-1));
               if(izq.cardinal()<Orden-1){ //Resuelto
                  rotacionderizq(padre,pos-1,izq,nodoActual);
                  return;
               }
            }
            if(pos < padre.cardinal()){ //Tiene hermano derecho
               der=leer(padre.enlace(pos+1));
               if(der.cardinal()<Orden-1){ //Resuelto
                  rotacionizqder(padre,pos,nodoActual,der);
                  return;
               }
            }
            //No se puede rotar => se parte
            if(pos ==0)
               particion_2_3(padre,pos,nodoActual,der);
            else
               particion_2_3(padre,pos-1,izq,nodoActual);
            if(padre.cardinal()< Orden){
               escribir(padre);
               return;
            }
            nodoActual=padre;
         }
         //Se parte la raz
         particion_1_2(nodoActual, claveInsertar, enlaceInsertar);
      }
      //Se crea una nueva raz
      nodoActual.cardinal(1);
      nodoActual.enlace(0)=this->raiz;
      nodoActual.clave(1)=claveInsertar;
      nodoActual.enlace(1)=enlaceInsertar;
      nodoActual.direccion(this->fichero.tomarPagina());
      this->raiz= nodoActual.direccion();
      escribir(nodoActual);
      this->fichero.adjunto(0,this->raiz);
   }
template <class T, int Orden>
   void ArbolB<T,Orden>::extraer(const T &dato){
      stack<iPila> pila;
      if(!buscar(dato, pila)) //Si no est salimos sin hacer nada
         return;
      this->fichero.adjunto(1,--this->numEle);
      Nodo nodoActual;
      iPila info = pila.top();
      nodoActual = info.nodo;
      int pos = info.pos;
      if(nodoActual.enlace(0) != dirNula){ //Extraccin desde un nodo no hoja
         //Hay que buscar el sucesor y cambiarlo
         queue<iPila> cola;
         int dir= nodoActual.enlace(pos);
         do{  //Descendemos por las ramas izquierda del subrbol del sucesor
            nodoActual=this->leer(dir);
            dir = nodoActual.enlace(0);
            if(dir== dirNula) pos=1;
            else pos=0;
            cola.push(iPila(nodoActual,pos)); //Guardamos el camino en una cola
         }while(dir != dirNula);
         info = pila.top();
         info.nodo.clave(info.pos)= nodoActual.clave(1);  //Se sustituye por el sucesor
         escribir(info.nodo); //Se escribe por si no hay ms modificaciones
         pila.pop();
         pila.push(info);
         while(!cola.empty()){
            pila.push(cola.front());
            cola.pop();
         }
      }
      //Extraccin en un nodo hoja
      nodoActual.extraer(pos);
      while(nodoActual.cardinal()< minimoclaves && nodoActual.direccion()!=this->raiz)
      {
         Nodo padre,der,izq;
         pila.pop();
         padre=pila.top().nodo;  //Se toma de la pila el padre
         pos = pila.top().pos;
         if(pos < padre.cardinal()) { //Tiene hermano derecho
            der=leer(padre.enlace(pos+1));
            if(der.cardinal()>minimoclaves){
               rotacionderizq(padre,pos,nodoActual,der);
               return;
            }
         }
         if(pos > 0) { //Tiene hermano izquierdo
            izq=leer(padre.enlace(pos-1));
            if(izq.cardinal()>minimoclaves){
               rotacionizqder(padre,pos-1,izq,nodoActual);
               return;
            }
         }
         //No se puede rotar => se recombina
         if(pos>0 && pos < padre.cardinal())
            recombinacion_3_2(padre,pos,izq,nodoActual,der);
         else if(pos>0)
            recombinacion_2_1(padre,pos-1,izq,nodoActual);
         else
            recombinacion_2_1(padre,pos,nodoActual,der);
         nodoActual=padre;
      }
      if(nodoActual.cardinal()>0){ //Se escribe el nodo, si tiene informacin
         escribir(nodoActual);
      }
      else //La raz se ha quedado sin datos
      {
         this->raiz = nodoActual.enlace(0);
         this->fichero.liberarPagina(nodoActual.direccion());
         this->fichero.adjunto(0,this->raiz);
      }
   }

#endif
