#ifndef INC_MULTIRRAMA
#define INC_MULTIRRAMA
#include <assert.h>
#include "..\comun\fichero.h"
#include "..\comun\contenedor.h"
#include <stack>
   using namespace std;

template <class T, int Orden>
   struct NodoMR{
      T   clavei[Orden];    //Claves   [1..numEle]
      int enlacei[Orden+1]; //Enlaces  [0..numEle]
      int numElei;          //Nmero de datos en la pgina
      int direccioni;       //Direccin de esta pgina en el fichero
   public:
      NodoMR(){
         direccioni=dirNula;
         enlacei[0]=dirNula;
         numElei=0;
      }
      T &clave(int i){ //Devuelve o establece la clave
         if(i<1 || i >numElei)
            throw excepcionContenedor("Error interno en multirrama");
         return clavei[i-1];}
      T clave(int i) const{ //Devuelve la clave
         if(i<1 || i >numElei)
            throw excepcionContenedor("Error interno en multirrama");
         return clavei[i-1];}
      int &enlace(int i){ //Devuelve o establece el enlace
         if(i<0 || i >numElei)
            throw excepcionContenedor("Error interno en multirrama");
         return enlacei[i];}
      int enlace(int i) const{//Devuelve el enlace
         if(i<0 || i >numElei)
            throw excepcionContenedor("Error interno en multirrama");
         return enlacei[i];}
      int cardinal() const{ //Devuelve el nmero de datos almacenados
         return numElei;}
      void cardinal(int n){ //Establece el nmero de datos almacenados
         if(n>numElei)
            for(int i=numElei+1; i<=n; i++) enlacei[i]=dirNula;
         numElei=n;
      }
      int direccion() const{ //Devuelve la direccin donde se alamacena
         return direccioni;}
      void direccion(int d){ //Establece la direccin donde se alamacena
         direccioni=d;}
      bool buscar(const T &e, int &pos)const;
      void insertar(const T &e, int dir, int pos);
      void extraer(int pos);
   };
	/**
    * Devuelve si lo ha encontrado o no y en pos la posicin
    * del la clave menor que es mayor o igual a e
    */
template <class T, int Orden>
   bool NodoMR<T,Orden>::buscar(const T &e, int &pos)const{
      int prim, ulti, med;
      prim = 1;
      ulti = cardinal();
      while(prim <= ulti){
         med =(prim+ulti)/2;
         if(e == clave(med)){
            pos = med;
            return true;
         }
         if(e < clave(med)) ulti = med-1;
         else               prim = med+1;
      }
      pos= prim-1;
      return false;
   }
template <class T, int Orden>
   void NodoMR<T,Orden>::insertar(const T &e, int dir, int pos){
      numElei++;
      for(int i=numElei-1; i>=pos; i--){
         clave(i+1)=clave(i);
         enlace(i+1)=enlace(i);
      }
      clave(pos)=e;
      enlace(pos)=dir;
   }
template <class T, int Orden>
   void NodoMR<T,Orden>::extraer( int pos){
      for(int i=pos; i<numElei; i++){
         clave(i)=clave(i+1);
         enlace(i)=enlace(i+1);
      }
      numElei--;
   }

template <class T, int Orden>
   struct infoPila{
      NodoMR<T, Orden> nodo;
      int pos;
      infoPila(){};
      infoPila(NodoMR<T, Orden> &n, int p):nodo(n),pos(p){}
   };

/**
 * Clase abstracta base de variantes del multirrama 
 */
template <class T, int Orden>
   class Multirrama: public ContenedorPersistente<T>{
   protected:
      int raiz;   //Direccion de la raiz en el fichero
      int numEle; //Nmero de datos en el multirrama
      typedef NodoMR<T,Orden> Nodo; //Nodo, alias para los nodosMR
      typedef infoPila<T,Orden> iPila;
      string nombreFichero;
      FicheroAvanzado<Nodo> fichero; //Fichero donde se almacenan los nodos
      void comprueba(const Nodo &)const; //Comprueba la corectitud del nodo
      Nodo leer(int dir) const;     //Escribe un nodo en el fichero
      void escribir(const Nodo &n); //Lee un nodo del fichero
      bool buscar(const T&e, stack<iPila> &pila) const;
   public:
      bool encuentra(const T&e) const;
      int  cardinal() const {return numEle;}
      virtual ~Multirrama() {};
   };

/**
 * Procedimiento que comprueba la correctitud del nodo pasado
 */
template <class T, int Orden>
   void Multirrama<T,Orden>::comprueba(const Nodo &n) const{
      if(n.cardinal() <= 0 || n.cardinal() >= Orden)
         throw excepcionContenedor("Error interno");
      bool esNodoHoja=false;
      for(int i=0; i<= n.cardinal(); i++){
         if(n.enlace(i)==dirNula)
            esNodoHoja=true;
      }
      if(esNodoHoja)
         for(int i=0; i<= n.cardinal(); i++){
            if(n.enlace(i)!=dirNula)
               throw excepcionContenedor("Error interno");
         }
      for(int i=1; i< n.cardinal(); i++)
         if(n.clave(i) >= n.clave(i+1))
            throw excepcionContenedor("Error interno");
   }

template <class T, int Orden>
   typename Multirrama<T,Orden>::Nodo Multirrama<T,Orden>::leer(int dir) const{
      Nodo n;
      n=fichero.leer(dir);
      if(n.direccion() != dir)
         throw excepcionContenedor("Error al leer un nodo del rbol");
      comprueba(n);
      return n;
   }
template <class T, int Orden>
   void Multirrama<T,Orden>::escribir(const Nodo &n) {
      comprueba(n);
      fichero.escribir(n,n.direccion());
   }
template <class T, int Orden>
   bool Multirrama<T,Orden>::buscar(const T&e, stack<iPila> &pila) const{
      int dirNodo, pos;
      Nodo nodo;
      bool encontrado;
      dirNodo = raiz;
      while(!pila.empty()) pila.pop();
      while(dirNodo != dirNula){
         nodo=leer(dirNodo);
         encontrado=nodo.buscar(e,pos);
         pila.push(iPila(nodo,pos));
         if(encontrado) 
            return true;
         dirNodo=nodo.enlace(pos);
      }
      return false;
   }

template <class T, int Orden>
   bool Multirrama<T,Orden>::encuentra(const T&e) const{
      stack<iPila> pila;
      return buscar(e,pila); 
   }

#endif
