#ifndef INC_FICHERO
#define INC_FICHERO
#include <cstdio>
#include <string>
#include <exception>
   using namespace std;
   class excepcionFichero: public exception{
      string s;
   public:
      excepcionFichero(string e){s=e;}
   	~excepcionFichero() throw() {}
      const char* what() const throw(){return s.c_str();}	
   };

   static const int dirNula=-1;
template <class T> 
   class Fichero{
      FILE *fichero;
      string nombre;
   public:
      Fichero(){
         fichero=NULL;}
      ~Fichero(){
         cerrar();}
      void crear(const string &);
      void abrir(const string &);
      void cerrar();
      int  tamano() const;
      bool abierto(){
         return fichero!=NULL;}
      T    leer(int) const;
      void escribir(const T &,int);
   };
template <class T> 
   void Fichero<T>::crear(const string &nombreFichero){
      cerrar();
      nombre=nombreFichero;
      fichero=fopen(nombre.c_str(),"w+b");
      if(fichero==NULL) 
         throw excepcionFichero(string("Error al crear el fichero "+nombre));
   }
template <class T> 
   void Fichero<T>::abrir(const string &nombreFichero){
      cerrar();
      nombre=nombreFichero;
      fichero=fopen(nombre.c_str(),"r+b");
      if(fichero==NULL) 
         throw excepcionFichero("Error al abrir el fichero "+nombre);
   }
template <class T> 
   void Fichero<T>::cerrar(){
      if(fichero!=NULL){  //FALTA COMPROBAR CONDICION DE ERROR
         if(fclose(fichero)) 
            throw excepcionFichero("Error al cerrar el fichero "+nombre);
         fichero=NULL;
      }
   }
template <class T> 
   int Fichero<T>::tamano() const{
      if(fichero!=NULL){
         fseek(fichero,0,SEEK_END); //FALTA COMPROBAR RESULTADO
         return ftell(fichero)/sizeof(T);
      }
      return 0;
   }
template <class T>
   T Fichero<T>::leer(int pos) const{
      if(fichero==NULL) 
         throw excepcionFichero("Intento de leer desde un fichero no abierto");
      if(fseek(fichero,pos*sizeof(T),SEEK_SET)) //FALTA COMPROBAR RESULTADO
         throw excepcionFichero("Error al posicionar el cursor en el fichero "+nombre);
      T dato;
      if(fread(&dato,sizeof(T),1,fichero) != 1)
         throw excepcionFichero("Error al leer datos del fichero "+nombre);
      return dato;
   }
template <class T> 
   void Fichero<T>::escribir(const T &dato,int pos){
      if(fichero==NULL) 
         throw excepcionFichero("Intento de escribir en un fichero no abierto");
      if(fseek(fichero,pos*sizeof(T),SEEK_SET)) //FALTA COMPROBAR RESULTADO
         throw excepcionFichero("Error al posicionar el cursor en el fichero "+nombre);
      if(fwrite(&dato,sizeof(T),1,fichero) != 1)
         throw excepcionFichero("Error al escribir datos en el fichero "+nombre);
   }

template <class T> 
   class FicheroAvanzado{
      FILE   *fichero;
      int    numeroAdjuntos;
      int    listaVacias;
      int    *adjuntos;   //Datos adjuntos al fichero
      int   desplazamiento;
      string nombre;
   	//Realizan las operaciones correspondientes sobre el fichero
      void   posicionar(int) const;  
      void   leerInterno(void *,int) const;
      void   escribirInterno(const void *,int);
   public:
      int    tomarPagina();
      void    liberarPagina(int );
   
      FicheroAvanzado(){
         desplazamiento=3*sizeof(int);
         numeroAdjuntos=0;
         adjuntos=NULL;
         fichero=NULL;
      }
      ~FicheroAvanzado(){
         cerrar();}
      void crear(const string &, int nAdjuntos=0);
      void abrir(const string &);
      int  adjunto(int );      //devuelve el valor del adjunto indicado
      void adjunto(int , int); //Modifica el valor del adjunto indicado
      void cerrar();
      int  tamano() const;
      bool abierto(){
         return fichero!=NULL;}
      int  nuevo();
      T    leer(int) const;
      void escribir(const T &,int);
   };
template <class T> 
   void   FicheroAvanzado<T>::posicionar(int pos)const{
      if(fseek(fichero,pos,SEEK_SET)) //FALTA COMPROBAR RESULTADO
         throw excepcionFichero("Error al posicionar el cursor en el fichero "+nombre);	
   } 
template <class T> 
   void   FicheroAvanzado<T>::leerInterno(void *dato,int t) const{
      if(fread(dato,t,1,fichero) != 1)
         throw excepcionFichero("Error al leer datos en el fichero "+nombre);
   }
template <class T> 
   void   FicheroAvanzado<T>::escribirInterno(const void *dato,int t){
      if(fwrite(dato,t,1,fichero) != 1)
         throw excepcionFichero("Error al escribir datos en el fichero "+nombre);
   }

template <class T> 
   void FicheroAvanzado<T>::crear(const string &nombreFichero, int nAdjuntos){
      cerrar();
      nombre=nombreFichero;
      fichero=fopen(nombre.c_str(),"w+b");
      if(fichero==NULL) 
         throw excepcionFichero(string("Error al crear el fichero "+nombre));
      numeroAdjuntos=nAdjuntos;
      int sizeT=sizeof(T);
      listaVacias=dirNula;
      escribirInterno(&sizeT,sizeof(int));   //Tamao de los registros
      escribirInterno(&numeroAdjuntos,sizeof(int));   //Nm. datos adjuntos
      escribirInterno(&listaVacias,sizeof(int));   //Lista pginas vacas
      if(numeroAdjuntos>0){
         adjuntos= new int[numeroAdjuntos];
         for(int i=0; i<numeroAdjuntos; i++)
            adjuntos[i]=0;
         escribirInterno(adjuntos, sizeof(int)*numeroAdjuntos);  //Adjuntos a cero
      }
      else adjuntos=NULL;
      desplazamiento=(3+numeroAdjuntos)*sizeof(int);
   }
template <class T> 
   void FicheroAvanzado<T>::abrir(const string &nombreFichero){
      cerrar();
      nombre=nombreFichero;
      fichero=fopen(nombre.c_str(),"r+b");
      if(fichero==NULL) 
         throw excepcionFichero("Error al abrir el fichero "+nombre);
      int sizeT;
      leerInterno(&sizeT, sizeof(int));   //Tamao de los registros
      if(sizeT != sizeof(T))
         throw excepcionFichero("El tipo del fichero "+nombre+" no es el correcto.");
      leerInterno(&numeroAdjuntos, sizeof(int)); //Nm. datos asociados
      leerInterno(&listaVacias, sizeof(int));    //Lista paginas vacas
      if(numeroAdjuntos>0){
         adjuntos= new int[numeroAdjuntos];
         leerInterno(adjuntos, sizeof(int)*numeroAdjuntos);  //Adjuntos a cero
      }
      else adjuntos=NULL;
      desplazamiento=(3+numeroAdjuntos)*sizeof(int);
   }
template <class T> 
   int FicheroAvanzado<T>::adjunto(int i){
      return adjuntos[i];
   }
template <class T> 
   void FicheroAvanzado<T>::adjunto(int i, int valor){
      adjuntos[i]=valor;
      posicionar((3+i)*sizeof(int));
      escribirInterno(&valor,sizeof(int));
   }
template <class T> 
   void FicheroAvanzado<T>::cerrar(){
      if(fichero!=NULL){  //FALTA COMPROBAR CONDICION DE ERROR
         if(fclose(fichero)) 
            throw excepcionFichero("Error al cerrar el fichero "+nombre);
         fichero=NULL;
      }
      if(adjuntos !=NULL){
         delete adjuntos;
         adjuntos=NULL;
      }
   }
template <class T> 
   int FicheroAvanzado<T>::tamano() const{
      if(fichero!=NULL){
         if(fseek(fichero,0,SEEK_END))
            throw excepcionFichero("Error posicionarse en el fichero "+nombre);
         return (ftell(fichero)-desplazamiento)/sizeof(T);
      }
      return 0;
   }
template <class T>
   int FicheroAvanzado<T>::tomarPagina(){
      if(listaVacias==dirNula) 
         return tamano();
      int pnueva=listaVacias;
      posicionar(desplazamiento+listaVacias*sizeof(T));
      leerInterno(&listaVacias,sizeof(int));
      posicionar(2*sizeof(int));
      escribirInterno(&listaVacias,sizeof(int));
      return pnueva;
   }
template <class T>
   void FicheroAvanzado<T>::liberarPagina(int pos){
      posicionar(desplazamiento+pos*sizeof(T));
      escribirInterno(&listaVacias,sizeof(int));
      listaVacias=pos;
      posicionar(2*sizeof(int));
      escribirInterno(&listaVacias,sizeof(int));
   }

template <class T>
   T FicheroAvanzado<T>::leer(int pos) const{
      T dato;
      posicionar(desplazamiento+pos*sizeof(T));
      leerInterno(&dato,sizeof(T));
      return dato;
   }
template <class T> 
   void FicheroAvanzado<T>::escribir(const T &dato,int pos){
      posicionar(desplazamiento+pos*sizeof(T));
      escribirInterno(&dato,sizeof(T));
   }

#endif

