package multirramas;

import comn.*;
import java.util.*;
import contenedores.AlmacenableComparable;
import contenedores.ContenedorPersistente;
import contenedores.ExcepcinContenedor;

/**
 * Clase abstracta base de variantes del multirrama 
 */
public abstract class Multirrama implements ContenedorPersistente {
   protected int raiz; //Direccion de la raiz en el fichero
   protected int numElem; //Nmero de datos en el multirrama
   protected int Orden; //Orden del rbol B
   protected int tamaoDatos; //Tamao de cada dato almacenado
   protected String nombreFichero;
   protected FicheroAvanzado fichero; //Fichero donde se almacenan los nodos
   protected AlmacenableComparable conv; //Objeto para conversiones a typo almacenado 

   protected class Nodo {
      private byte[][] clavei; //Claves [1..numEle]
      private int[] enlacei; //Enlaces [0..numEle]
      private int numElei; //Nmero de datos en la pgina
      private int direccini; //Direccion de la pgina en el fichero

      Nodo() {
         direccini = Fichero.dirNula;
         numElei = 0;
         //El tamao depende de "Orden" del objeto donde se crea
         clavei = new byte[Orden][];  
         enlacei = new int[Orden + 1];
      }

      int tamao() {
         int tam = 2 * Conversor.INTBYTES; //Cantidad base
         tam += (Orden - 1) * tamaoDatos; //Los datos
         tam += Orden * Conversor.INTBYTES;//Los enlaces
         return tam;
      }

      byte[] clave(int i) { //Devuelve la clave
         if (i < 1 || i > numElei)
            throw new ExcepcinContenedor("Error interno en multirrama");
         return clavei[i - 1];
      }

      void clave(int i, byte[] d) { //establece la clave
         if (i < 1 || i > numElei)
            throw new ExcepcinContenedor("Error interno en multirrama");
         clavei[i - 1] = d;
      }

      int enlace(int i) { //Devuelve el enlace
         if (i < 0 || i > numElei)
            throw new ExcepcinContenedor("Error interno en multirrama");
         return enlacei[i];
      }

      void enlace(int i, int d) { //Establece el enlace
         if (i < 0 || i > numElei)
            throw new ExcepcinContenedor("Error interno en multirrama");
         enlacei[i] = d;
      }

      int direccin() { //Devuelve la direccin donde se alamacena
         return direccini;
      }

      void direccin(int d) { //Establece la direccin donde se alamacena
         direccini = d;
      }

      int cardinal() { //Devuelve el nmero de datos almacenados
         return numElei;
      }

      void cardinal(int n) { //Establece el nmero de datos almacenados
         numElei = n;
      }

      /**
       * @return devuelve el nodo en forma de byte[]
       */
      byte[] aByte() {
         int tam = tamao();
         byte[] res = new byte[tam];
         int pos = 0;
         pos = Conversor.aade(res, Conversor.aByte(direccini), pos);
         pos = Conversor.aade(res, Conversor.aByte(numElei), pos);
         for (int i = 0; i < numElei; i++)
            pos = Conversor.aade(res, clavei[i], pos);
         for (int i = 0; i <= numElei; i++)
            pos = Conversor.aade(res, Conversor.aByte(enlacei[i]), pos);
         return res;
      }

      /**
       * @param datos
       * Inicializa el nodo partiendo de un byte[] previamente generado
       * con aByte
       */
      void deByte(byte[] datos) {
         int leb = Conversor.INTBYTES; //Longitud de los enteros en bytes
         direccin(Conversor.aInt(Conversor.toma(datos, 0, leb)));
         numElei = Conversor.aInt(Conversor.toma(datos, leb, leb));
         int baseClaves = leb * 2;
         int baseEnlaces = baseClaves + (numElei) * tamaoDatos;
         for (int i = 0; i < numElei; i++)
            clavei[i] = Conversor.toma(datos, baseClaves + i * tamaoDatos,
                  tamaoDatos);
         for (int i = 0; i <= numElei; i++) {
            byte[] dato = Conversor.toma(datos, baseEnlaces + i * leb, leb);
            enlacei[i] = Conversor.aInt(dato);
         }
      }

      /**
       * @param e dato a buscar en forma de byte[]
       * @return devuelve la posicin del la clave menor que es mayor o
       * igual a e.
       */
      public int buscar(AlmacenableComparable e) {
         int pos, prim, ulti, med;
         prim = 1;
         ulti = cardinal();
         while (prim <= ulti) {
            med = (prim + ulti) / 2;
            if (e.compareTo(e.deByte(clave(med))) == 0) {
               pos = med;
               return pos;
            }
            if (e.compareTo(e.deByte(clave(med))) < 0) ulti = med - 1;
            else prim = med + 1;
         }
         pos = prim - 1;
         return pos;
      }

      /**
       * @param e
       * dato a buscar en forma de byte[]
       * @return devuelve si el dato pasado est o no el nodo.
       */
      public boolean est(AlmacenableComparable e) {
         int prim, ulti, med;
         prim = 1;
         ulti = cardinal();
         while (prim <= ulti) {
            med = (prim + ulti) / 2;
            if (e.compareTo(e.deByte(clave(med))) == 0) {
               return true;
            }
            if (e.compareTo(e.deByte(clave(med))) < 0) ulti = med - 1;
            else prim = med + 1;
         }
         return false;
      }

      void insertar(byte[] e, int dir, int pos) {
         numElei++;
         for (int i = numElei - 1; i >= pos; i--) {
            clave(i + 1, clave(i));
            enlace(i + 1, enlace(i));
         }
         clave(pos, e);
         enlace(pos, dir);
      }

      void extraer(int pos) {
         for (int i = pos; i < numElei; i++) {
            clave(i, clave(i + 1));
            enlace(i, enlace(i + 1));
         }
         numElei--;
      }
   }

   class InfoPila {
      public Nodo nodo;
      public int pos;

      public InfoPila() {
      };

      public InfoPila(Nodo n, int p) {
         nodo = n;
         pos = p;
      }
   }

   int cardinal() {
      return numElem;
   }

   /**
    * Procedimiento que comprueba la correctitud del nodo pasado
    */
   void comprueba(Nodo n) {
      if (n.cardinal() <= 0 || n.cardinal() >= Orden)
         throw new ExcepcinContenedor("Error interno");
      boolean esNodoHoja = false;
      for (int i = 0; i <= n.cardinal(); i++) {
         if (n.enlace(i) == Fichero.dirNula) esNodoHoja = true;
      }
      if (esNodoHoja) for (int i = 0; i <= n.cardinal(); i++) {
         if (n.enlace(i) != Fichero.dirNula)
            throw new ExcepcinContenedor("Error interno");
      }
      for (int i = 1; i < n.cardinal(); i++) {
         AlmacenableComparable a =
            (AlmacenableComparable) conv.deByte(n.clave(i));
         AlmacenableComparable b =
            (AlmacenableComparable) conv.deByte(n.clave(i + 1));
         if (a.compareTo(b) >= 0)
            throw new ExcepcinContenedor("Error interno");
      }
   }

   public Multirrama() {
      fichero = new FicheroAvanzado();
   }

   Nodo leer(int dir) {
      Nodo n = new Nodo();
      n.deByte(fichero.leer(dir));
      if (n.direccin() != dir)
         throw new ExcepcinContenedor("Error al leer un nodo del rbol");
      comprueba(n);
      return n;
   }

   void escribir(Nodo n) {
      comprueba(n);
      fichero.escribir(n.aByte(), n.direccin());
   }

   /**
    * @param e dato a buscar en forma de byte[]
    * @param pila donde se almacena el camino 
    * @return si lo ha encontrado o no
    */
   boolean buscar(AlmacenableComparable e, Stack pila) {
      int dirNodo, pos;
      Nodo nodo = new Nodo();
      dirNodo = raiz;
      pila.clear();
      while (dirNodo != Fichero.dirNula) {
         nodo = leer(dirNodo);
         pos = nodo.buscar(e);
         pila.add(new InfoPila(nodo, pos));
         if (nodo.est(e)) return true;
         dirNodo = nodo.enlace(pos);
      }
      return false;
   }

   boolean encuentra(AlmacenableComparable e) {
      Stack pila = new Stack();
      return buscar(e, pila);
   }
}